@@ -22,6 +22,16 @@
 #define PATH_MAILLOG	"/var/adm/syslog/mail.log"
 #endif
 
+#ifdef SUNOS
+#define HAVE_GETRUSAGE
+#define PATH_TMP        "/tmp"
+#define PATH_VAR_TMP    "/var/tmp"
+#define PATH_PASSWD     "/etc/passwd"
+#define PATH_WTMP       "/var/adm/wtmp"
+#define PATH_UTMP       "/etc/utmp"
+#define PATH_SYSLOG     "/var/log/syslog"
+#endif
+
 #ifndef RANDSAVENAME
 #define RANDSAVENAME	"/etc/prngd-seed"
 #endif
