@@ -18,12 +18,12 @@
 #include "config.h"
 #include "prngd.h"
 
-static int gatherer_pid = -1;
+static pid_t gatherer_pid = -1;
 
 typedef enum
 {
   wait_command, wait_second_byte, wait_third_byte, wait_fourth_byte,
-  wait_read, wait_write, to_be_closed
+  wait_read, wait_write, wait_entropy, to_be_closed
 }
 querystate_t;
 
@@ -169,46 +169,51 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
     }
 
     if (gather_fd != -1) {
-    if (FD_ISSET(gather_fd, &efds)) {
-      close(gather_fd);
-      gather_fd = -1;
-    }
-    if (FD_ISSET(gather_fd, &rfds)) {
-      /*
-       * Bytes are coming back from the gatherer process. Read chunks of
-       * up to 10240 bytes and feed them into the PRNG
-       */
-      numbytes = read(gather_fd, gatherbuf, 10240);
-      rand_add(gatherbuf, numbytes, est * numbytes);
-      if (debug && numbytes > 0)
-	fprintf(stderr, "Read %d bytes from gatherer\n", numbytes);
-      gatherer_bytes += numbytes;
-      if ((numbytes <= 0) || (gatherer_bytes >= max_gatherer_bytes)) {
-	if (debug)
-	  fprintf(stderr, "Closing connection to gatherer\n");
+      if (FD_ISSET(gather_fd, &efds)) {
 	close(gather_fd);
 	gather_fd = -1;
-	gatherer_entry++;
-        if (entropy_source[gatherer_entry].path == NULL) {
-	  first_gather = 0;	/* done */
-	  gatherer_entry = 0;
+      }
+      else if (FD_ISSET(gather_fd, &rfds)) {
+	/*
+	 * Bytes are coming back from the gatherer process. Read chunks of
+	 * up to 10240 bytes and feed them into the PRNG
+	 */
+	numbytes = read(gather_fd, gatherbuf, 10240);
+	rand_add(gatherbuf, numbytes, est * numbytes);
+	if (debug && numbytes > 0)
+	  fprintf(stderr, "Read %d bytes from gatherer\n", numbytes);
+	gatherer_bytes += numbytes;
+	if ((numbytes <= 0) || (gatherer_bytes >= max_gatherer_bytes)) {
+	  if (debug)
+	    fprintf(stderr, "Closing connection to gatherer\n");
+	  close(gather_fd);
+	  gather_fd = -1;
+	  gatherer_entry++;
+          if (entropy_source[gatherer_entry].path == NULL) {
+	    first_gather = 0;	/* done */
+	    gatherer_entry = 0;
+	  }
 	}
       }
     }
-    }
 
     /*
      * If a gatherer has not finished in time, kill it!!
      */
     if ((gatherer_pid > 0) &&
 	(difftime(time(NULL), gather_start) > seed_ext_interval / 2 + 1)) {
-	kill(gatherer_pid, SIGKILL);
-	gatherer_pid = -1;
-	close(gather_fd);
-	gather_fd = -1;
+      kill(gatherer_pid, SIGKILL);
+      gatherer_pid = -1;
+      close(gather_fd);
+      gather_fd = -1;
     }
 
-    if ((time(0) >= last_seed_ext + seed_ext_interval) && (gather_fd == -1))
+    /*
+     * Call a gatherer, if it is time to do so. When starting up, call
+     * all gatherers once to obtain initial seeding.
+     */
+    if (((time(0) >= last_seed_ext + seed_ext_interval) || first_gather) &&
+	(gather_fd == -1))
     {
       last_seed_ext = time(0);
       /*
@@ -296,9 +301,14 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
 	egd_query[i].state = to_be_closed;
       }
       else if (FD_ISSET(egd_query[i].fd, &rfds) ||
-	       FD_ISSET(egd_query[i].fd, &wfds)) {
+	       FD_ISSET(egd_query[i].fd, &wfds) ||
+	       (egd_query[i].state == wait_entropy)) {
       switch (egd_query[i].state)
       {
+      case wait_entropy:
+	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0)
+	    egd_query[i].state = wait_write;
+	break;
       case wait_command:
 	error = read(egd_query[i].fd, &(egd_query[i].command), 1);
 	if ((error == -1) && (errno == EAGAIN))
@@ -309,10 +319,11 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
 	  egd_query[i].state = to_be_closed;
 	else if (egd_query[i].command == 0)
 	{
-	  egd_query[i].buffer[0] = 0;
-	  egd_query[i].buffer[1] = 0;
-	  egd_query[i].buffer[2] = 8192 / 256;
-	  egd_query[i].buffer[3] = 8192 % 256;
+	  numbytes = rand_pool();
+	  egd_query[i].buffer[0] = numbytes / 16777216;
+	  egd_query[i].buffer[1] = (numbytes % 16777216) / 65536;
+	  egd_query[i].buffer[2] = (numbytes % 65536) / 256;
+	  egd_query[i].buffer[3] = numbytes % 256;
 	  egd_query[i].to_process = 4;
 	  egd_query[i].yet_processed = 0;
 	  egd_query[i].state = wait_write;
@@ -336,18 +347,23 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
 	  egd_query[i].state = to_be_closed;
 	else if (egd_query[i].command == 1)
 	{
-	  egd_query[i].buffer[0] = egd_query[i].secbyte;
-	  rand_bytes(egd_query[i].buffer + 1, egd_query[i].secbyte);
-	  egd_query[i].to_process = egd_query[i].secbyte + 1;
+	  if (rand_bytes(egd_query[i].buffer + 1, egd_query[i].secbyte))
+	    numbytes = egd_query[i].secbyte;
+	  else
+	    numbytes = 0;
+	  egd_query[i].buffer[0] = numbytes;
+	  egd_query[i].to_process = numbytes + 1;
 	  egd_query[i].yet_processed = 0;
 	  egd_query[i].state = wait_write;
 	}
 	else if (egd_query[i].command == 2)
 	{
-	  rand_bytes(egd_query[i].buffer, egd_query[i].secbyte);
 	  egd_query[i].to_process = egd_query[i].secbyte;
 	  egd_query[i].yet_processed = 0;
-	  egd_query[i].state = wait_write;
+	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0)
+	    egd_query[i].state = wait_write;
+	  else
+	    egd_query[i].state = wait_entropy;
 	}
 	else if (egd_query[i].command == 3)
 	{
@@ -429,7 +445,8 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
     if (FD_ISSET(service_socket, &rfds))
     {
       sockunlen = sizeof(sockun);
-      accept_fd = accept(service_socket, &sockun, &sockunlen);
+      accept_fd = accept(service_socket, (struct sockaddr *) &sockun,
+			 &sockunlen);
       if (accept_fd < 0)
 	continue;		/*
 				 * Ignore and stay alive 
@@ -441,6 +458,7 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
 	  fprintf(stderr, "Could not get socket file mode: %s\n",
 		  strerror(errno));
 	close(accept_fd);
+	continue;
       }
       if (fcntl(accept_fd, F_SETFL, flags | O_NONBLOCK) < 0)
       {
@@ -448,6 +466,7 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
 	  fprintf(stderr, "Could not set socket file mode: %s\n",
 		  strerror(errno));
 	close(accept_fd);
+	continue;
       }
 
       /*
@@ -468,10 +487,15 @@ void main_loop(int service_socket, struct sockaddr_un sockun,
 				 * Better this than complete shut down 
 				 */
 	  egd_query = egd_query_old;
+	  continue;
 	}
+	alloc_query = max_query + 1;
       }
+      /*
+       * This routine is safe here, as in case of error we do "continue;"
+       * in case of problems. So we have succesfully reached this point.
+       */
       max_query++;
-      alloc_query = max_query;
       egd_query[max_query - 1].fd = accept_fd;
       egd_query[max_query - 1].state = wait_command;
       fprintf(stderr, "Opening service fd %d\n", egd_query[max_query-1].fd);
